import * as METADATA from '../constants';
import { POUCH_DEFAULT_QUALIFIER } from '../constants';
import InjectConfig from './InjectConfig';
import PouchInjectDefinition from '../config/PouchInjectDefinition';

const Inject = (injectConfig: InjectConfig = {}) => {
  return (...args: any[]) => {
    let params = [];
    for (let i = 0; i < args.length; i++) {
      if (args[i]) {
        params.push(args[i]);
      }
    }

    switch(params.length) {
    case 1:
      return injectClass(injectConfig).apply(this, args);
    case 2:
      return injectProperty(injectConfig).apply(this, args);
    case 3:
      if(typeof args[2] === 'number') {
        return injectParameter(injectConfig).apply(this, args);
      }
      return injectMethod(injectConfig).apply(this, args);
    default:
      throw new Error('@Inject decorators are not valid here!');
    }
  };
};

const injectClass = (injectConfig: InjectConfig = {}) => {
  return <TFunction extends Function>(target: TFunction): TFunction | void => {
      throw new Error('Cannot apply @Inject decorator on Class.');
  };
};

const injectProperty = (injectConfig: InjectConfig = {}) => {
  return (target: Object, propertyKey: string | symbol): void => {
    let pouchInjectDefinition: PouchInjectDefinition = getPouchInjectDefinition(target);
    pouchInjectDefinition.addInject(injectConfig, propertyKey);
  };
};

const injectParameter = (injectConfig: InjectConfig = {}) => {
  return (target: Object, propertyKey: string | symbol, parameterIndex: number): void => {
    let pouchInjectDefinition: PouchInjectDefinition = getPouchInjectDefinition(target);
    pouchInjectDefinition.addInject(injectConfig, propertyKey, parameterIndex);
  };
};

const injectMethod = (injectConfig: InjectConfig = {}) => {
  return (target: Object, propertyKey: string | symbol,
          descriptor: TypedPropertyDescriptor<any>): TypedPropertyDescriptor<any> | void => {
    throw new Error('Cannot apply @Inject decorator on Class.');

    // return descriptor;
  };
};



const getPouchInjectDefinition = (target: Object) => {
    let pouchInjectDefinition: PouchInjectDefinition;
    if (Reflect.hasOwnMetadata(METADATA.POUCH_INJECT_DEFINITION, target) !== true) {
      pouchInjectDefinition = new PouchInjectDefinition(target);
      Reflect.defineMetadata(METADATA.POUCH_INJECT_DEFINITION, pouchInjectDefinition, target);
    } else {
      pouchInjectDefinition = Reflect.getMetadata(METADATA.POUCH_INJECT_DEFINITION, target);
    }
    return pouchInjectDefinition;
};

export default Inject;
