import * as METADATA from '../constants';
import PouchDefinition from '../config/PouchDefinition';


const Injectable = (...args: any[]) => {
  let params = [];
  for (let i = 0; i < args.length; i++) {
    if (args[i]) {
      params.push(args[i]);
    }
  }

  switch(params.length) {
  case 1:
    return injectableClass.apply(this, args);
  case 2:
    return injectableProperty.apply(this, args);
  case 3:
    if(typeof args[2] === 'number') {
      return injectableParameter.apply(this, args);
    }
    return injectableMethod.apply(this, args);
  default:
    throw new Error('@Injectable decorators are not valid here!');
  }
};


const injectableClass = <TFunction extends Function>(target: TFunction): TFunction | void => {
  if (Reflect.hasOwnMetadata(METADATA.POUCH_DEFINITION, target) === true) {
    throw new Error('Cannot apply @injectable decorator multiple times.');
  }

  let pouchDefinition: PouchDefinition = getPouchDefinition(target.prototype);

  return target;
};


const injectableProperty = (target: Object, propertyKey: string | symbol): void => {
  throw new Error('Cannot apply @Injectable decorator on property.');
};


const injectableParameter = (target: Object, propertyKey: string | symbol, parameterIndex: number): void => {
  throw new Error('Cannot apply @Injectable decorator on parameter.');
};

const injectableMethod = <T>(target: Object, propertyKey: string | symbol,
                             descriptor: TypedPropertyDescriptor<T>): TypedPropertyDescriptor<T> | void => {
  throw new Error('Cannot apply @Injectable decorator on method.');
  // return descriptor;
};

const getPouchDefinition = <TFunction extends Function>(target: TFunction) => {
    let pouchDefinition: PouchDefinition;
    if (Reflect.hasOwnMetadata(METADATA.POUCH_DEFINITION, target) !== true) {
      pouchDefinition = new PouchDefinition(target);
      Reflect.defineMetadata(METADATA.POUCH_DEFINITION, pouchDefinition, target);
    } else {
      pouchDefinition = Reflect.getMetadata(METADATA.POUCH_DEFINITION, target);
    }
    return pouchDefinition;
};


export default Injectable;
